runtime! syntax/javascript.vim

syn keyword javaScriptGlobal global session settings
syn keyword javaScriptMember clipboard data extensions gui io net script signals system tabs timer util
syn keyword javaScriptType Deferred Signal Glob
